import unittest
import zope.component
import Products.Five
from pprint import pprint
from interact import interact
from zope.testing import doctest
from zope.app.testing.placelesssetup import setUp, tearDown
from zope.configuration.xmlconfig import XMLConfig

optionflags = doctest.NORMALIZE_WHITESPACE | \
              doctest.ELLIPSIS | \
              doctest.REPORT_ONLY_FIRST_FAILURE

TESTFILES = [
    '../soup.txt',
]

def test_suite():
    setUp()
    XMLConfig('meta.zcml', zope.component)()
    XMLConfig('meta.zcml', Products.Five)()
    XMLConfig('configure.zcml', Products.Five)()
    return unittest.TestSuite([
        doctest.DocFileSuite(
            file, 
            optionflags=optionflags,
            globs={'interact': interact,
                   'pprint': pprint},
        ) for file in TESTFILES
    ])
    tearDown()

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite') 


from zope.interface import Interface
from zope.interface import Attribute

class ISoup(Interface):
    """The Container Interface.
    """
    
    id = Attribute(u"The id of this Soup")
    
    nextrecordindex = Attribute(u"The next record index to use.")

    def add(record):
        """Add record to soup.
        
        @param record: IRecord implementation
        @return: intid for record
        """
                    
    def query(**kw):
        """Query Soup for Records.
        
        @param kw: Keyword arguments defining the query
        @return: list of records
        """
        
    def rebuild(self):
        """replaces the catalog and reindex all records."""
    
    def reindex(record=None):
        """Reindex the catalog for this soup.
        
        if record is None reindex all records, otherwise a list of records is
        expected. 
        """
    
    def __delitem__(record):
        """Delete Record from soup.
        
        If given record not contained in soup, raise ValueError.
        
        @param record: IRecord implementation
        @raise: ValueError if record not exists in this soup.
        """

class IRecord(Interface):
    """The record Interface.
    """
    
    id = Attribute(u"The id of this Record")
    
    intid = Attribute(u"The intid of this record")
    
    data = Attribute(u"Dict like object representing the Record Data")

class ICatalogFactory(Interface):
    """Factory for the catalog used for Soup.
    """
    
    def __call__():
        """Create and return the Catalog.
        
        @param return: zope.app.catalog.catalog.Catalog instance
        """


import transaction

from zope.interface import implements
from zope.component import adapts
from zope.component import getUtility
from zope.location import Location
from zope.app.intid.interfaces import IIntIds
from zope.container.interfaces import INameChooser
from zope.keyreference.interfaces import IKeyReference, NotYet

from BTrees.OOBTree import OOBTree
from OFS.SimpleItem import SimpleItem
from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2

from interfaces import ISoup
from interfaces import IRecord
from interfaces import ICatalogFactory

class KeyReferenceToRecord(object):
    """IKeyReference implementation referencing Records inside a Soup.
    """
    implements(IKeyReference)
    adapts(IRecord)
    
    key_type_id = 'cornerstone.soup.keyreference'
    
    def __init__(self, wrapped_obj):
        if not wrapped_obj.id:
            raise NotYet()
        self.soupid = wrapped_obj.__parent__.id
        self.refid = wrapped_obj.id
    
    @property
    def storage(self):
        if not hasattr(self, '_storage'):
            self._storage = getUtility(ISoup, name=self.soupid)
        return self._storage
    
    def __call__(self):
        return self.storage.get(self.refid, None)

    def __hash__(self):
        return hash((self.soupid, self.storage, self.refid))

    def __cmp__(self, other):
        if self.key_type_id == other.key_type_id:
            myident = '%s.%s' % (self.soupid, self.refid)
            otherident = '%s.%s' % (other.soupid, other.refid)
            return cmp(myident, otherident)
        # XXX: check wether below comparsion is even
        return cmp(self.key_type_id, other.key_type_id)

class NameChooser(object):
    """INameChooser implementation for records.
    """
    implements(INameChooser)
    
    def __init__(self, context):
        self.context = context

    def checkName(self, name, object):
        if not name == 'record':
            raise ValueError(u"Odd name given")

    def chooseName(self, name, object):
        self.checkName(name, object)
        name = '%s-%i' % (name, self.context.nextrecordindex)
        self.context.nextrecordindex += 1
        return name

class Soup(BTreeFolder2, Location):
    """ISoup implementation.
    
    Abstract. This class is supposed to be derived from.
    """
    implements(ISoup)
    
    id = None
    
    def __init__(self, *args, **kw):
        BTreeFolder2.__init__(self, *args, **kw)
        self.nextrecordindex = 0
    
    @property
    def catalog(self):
        if not hasattr(self, '_catalog'):
            self._catalog = getUtility(ICatalogFactory, name=self.id)()
        return self._catalog

    def add(self, record):
        nc = INameChooser(self)
        id = nc.chooseName('record', record)
        record.id = id
        record.__parent__ = self
        self[id] = record
        record = self[id]
        intids = getUtility(IIntIds, name=self.id)
        record.intid = intids.register(record)
        self.catalog.index_doc(record.intid, record)
        # XXX: notify subscribers here
        return record.intid
                    
    def query(self, **kw):
        querykw = {}
        for key in kw:
            if isinstance(kw[key], list) \
              or isinstance(kw[key], tuple):
                assert(len(kw[key]) == 2)
                querykw[key] = kw[key]
            else:
                querykw[key] = (kw[key], kw[key])
        ids = self.catalog.apply(querykw)
        intids = getUtility(IIntIds, name=self.id)
        result = list()
        for id in ids:
            record = intids.getObject(id)
            result.append(record)
        return result
    
    def rebuild(self):
        self._catalog = getUtility(ICatalogFactory, name=self.id)()
        self.reindex()
    
    def reindex(self, records=None):
        if records is None:
            records = self.objectValues()
        for record in records:
            if record.intid is None:
                intids = getUtility(IIntIds, name=self.id)   
                intids.unregister(record) # paranoid
                record.intid = intids.register(record)
            self.catalog.index_doc(record.intid, record)

    def __delitem__(self, record):
        if not record.__parent__.id == self.id:
            raise ValueError(u"Record not contained in this soup")
        BTreeFolder2.__delitem__(self, record.id)
        self.catalog.unindex_doc(record.intid)

EMPTY_MARKER = object()

class Record(SimpleItem, Location):
    """IRecord implementation.
    """
    implements(IRecord)
    
    def __init__(self, **kw):
        self.id = None
        self.intid = None
        self.data = OOBTree()
        for key in kw.keys():
            self.data[key] = kw[key]
        self._p_changed = True
    
    def __getattribute__(self, name):
        try:
            attr = SimpleItem.__getattribute__(self, 'data').get(name,
                                                                 EMPTY_MARKER)
            if attr is not EMPTY_MARKER:
                return attr
        except AttributeError, e: pass
        return SimpleItem.__getattribute__(self, name)


import code
import sys

def interact(locals=None):
    """Provides an interactive shell aka console inside your testcase.
    
    It looks exact like in a doctestcase and you can copy and paste
    code from the shell into your doctest. The locals in the testcase are 
    available, because you are _in_ the testcase.

    In your testcase or doctest you can invoke the shell at any point by
    calling::
        
        >>> interact( locals() )        
        
    locals -- passed to InteractiveInterpreter.__init__()
    """
    savestdout = sys.stdout
    sys.stdout = sys.stderr
    sys.stderr.write('\n'+'='*75)
    console = code.InteractiveConsole(locals)
    console.interact("""
DocTest Interactive Console - (c) BlueDynamics Alliance, Austria, 2006-2008
Note: You have the same locals available as in your test-case. 
Ctrl-D ends session and continues testing.
""")
    sys.stdout.write('\nend of DocTest Interactive Console session\n')
    sys.stdout.write('='*75+'\n')
    sys.stdout = savestdout 



from zope.interface import implements

from zope.catalog.catalog import Catalog
from zope.catalog.field import FieldIndex

from cornerstone.soup.interfaces import ICatalogFactory
from cornerstone.soup import Soup

class MySoup(Soup):
    id = u'mysoup'

class MyCatalogFactory(object):
    """ICatalogFactory implementation used for testing.
    """

    implements(ICatalogFactory)

    def __call__(self):
        catalog = Catalog()
        catalog[u'user'] = FieldIndex(field_name='user',
                                      field_callable=False)
        return catalog


from setuptools import setup, find_packages
import sys, os

version = '1.0b1'
shortdesc ="Container for queryable Records."
longdesc = open(os.path.join(os.path.dirname(__file__), 'README.txt')).read() 
longdesc += open(os.path.join(os.path.dirname(__file__), 'LICENSE.txt')).read()

setup(name='cornerstone.soup',
      version=version,
      description=shortdesc,
      long_description=longdesc,
      classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Framework :: Zope2',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Libraries :: Python Modules'        
      ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='container data record catalog',
      author='BlueDynamics Alliance',
      author_email='dev@bluedynamics.com',
      url='http://dev.plone.org/collective/browser/cornerstone.soup',
      license='BSD',
      packages=find_packages('src'),
      package_dir = {'': 'src'},
      namespace_packages=['cornerstone'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'zope.catalog',
          'five.intid',
          #'ZODB',
          # -*- Extra requirements: -*
      ],
      extras_require={
      },
      entry_points="""
      # -*- Entry points: -*-
      """,
      )